% :- set_dbg( off ). 
% :- set_dbg( on ).

:- restore( demo ).

%%% This file containing some small Pfd predicate examples.

%%% Predicates that have appeared in some paper.

% ???
/* 
 plucky( P ).
 P = 1/3. */

plucky( PLucky1 ) :-
	Die pin uniform([1,2,3,4,5,6]),
	LCoin pin finite_geometric([head,tail],2),
	PLucky1 is p(lucky(Die,LCoin)).


/*
fdm_plucky( PLucky1, PLucky2 ).
PLucky1=1/3, PLucky2=4/15.
*/
fdm_plucky( PLucky1, PLucky2 ) :-
	Die fpin uniform([1,2,3,4,5,6]),
	LCoin pin finite_geometric([head,tail],2),
	PLucky1 is p(lucky(Die,LCoin)),
	V in 4..5, 
	%
	V #\= Die, 
	V #\= 4,
	PLucky2 is p(lucky(Die,LCoin)).

/* 
This has nt appeared, it simply tests the interaction
with clpfd for non arithmetic domains. PLucky2 is 1/2
because LCoin == head, thus is = to the prb of Die = {4,5,6}.
coin_fdm_plucky( PLucky1, PLucky2 ).
PLucky1=1/3, PLucky2=1/2 ?  */
coin_fdm_plucky( PLucky1, PLucky2 ) :-
	Die pin uniform([1,2,3,4,5,6]),
	LCoin fpin finite_geometric([head,tail],2),
	PLucky1 is p(lucky(Die,LCoin)),
	V in 2..3, 
	%
	V #\= 3, 
	LCoin #\= V,
	PLucky2 is p(lucky(Die,LCoin)).

lucky( D, head ) :- 
	( D==4; D==5; D==6).

%%% Predicates illustrating abbreviation known to Pfd.
/*
no_pi( a, P ).
no_pi( b, P ).
no_pi( c, P ).
a = 1/2, b = c = 1/4.  */
no_pi( E, P ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	X = a \\ Y = x,
	P is p(X==E).
	% P(X==a) = 1/4 * 1 + 3*(1/4 * 1/3) = 1/2
	% P(X==b) = P(X==c) = 1/4 * 0 + 3*(1/4 * 1/3) = 1/4.
no_pi_distro( Distro ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	X = a \\ Y = x,
	static_distribution( X, Distro ).
/* 
no_pi_base( a, P ).
no_pi_base( b, P ).
no_pi_base( c, P ).
a = 1/2, b = c = 1/4.  */
no_pi_base( E, P ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	X = a :: 1/1 \\ Y = x,
	P is p(X==E).
	% compare to bellow 

/* 
rhs_pi( a, 1/2, P ).
rhs_pi( b, 1/2, P ).
rhs_pi( c, 1/2, P ).
a = 3/8, b = c = 5/16.
rhs_pi( a, 1/3, P ).
rhs_pi( a, 1/3, P ).
a = b = 1/3. */
rhs_pi( E, With, P ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	X = a \\ With :: Y = x,
	P is p(X==E).
	% P(X==a)= 1/4*With*1 + 3*(1/4*1/3) 
	% P(X==b)= P(X==c)= 1/4*(1 - With)*1/2 + 3*(1/4*1/3)
	% With = 1/2 
	% P(X==a)= 1/8 + 1/4 = 3/8
	% P(X==b)= P(X==c)= 1/16 + 1/4 = 5/16
	% With = 1/3 
	% P(X==a)= 1/12 + 1/4 = 1/3
	% P(X==b)= P(X==c)= 1/12 + 1/4 = 1/3
rhs_pi_base( E, With, P ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	% adding the following twice will, correctly, result in different distro.
	X = a :: With ++ (\+ X = a) :: (1 - With) \\ Y = x,
	P is p(X==E).

% in contrast to the above, the following looses 1/8 prb mass,
% 1/16 for each b and c.
/* 
mass_loss_rhs_pi( a, 1/2, P ).
mass_loss_rhs_pi( b, 1/2, P ).
mass_loss_rhs_pi( c, 1/2, P ).
a = 3/8, b = c = 1/4.  */
mass_loss_single_pi( E, With, P ) :-
	X pinl uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	X = a :: With \\ Y = x,
	P is p(X==E).

/* 
multi_no_loss( a, 1/2, P ).
multi_no_loss( b, 1/2, P ).
multi_no_loss( c, 1/2, P ).
a = b = c = 1/3.  */
multi_no_loss( E, With, P ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	X = a \\ With :: Y = x,
	X = b \\ With :: Y = x,
	P is p(X==E).
	% With is 1/2 
	% P(X==a)=P(X=b)=P(X=c)=1/3
	% (Dirived differently for a and b in comparison to c.)

% in contrast to the above, this uses a mass loss allowing declaration.
% With = 1/2, P(a)=P(b)=P(c)=5/16
% Loosing mass when a=X /\ b=X; phi applied to {a,b,c} above.
/* 
multi_with_loss( a, 1/2, P ).
multi_with_loss( b, 1/2, P ).
multi_with_loss( c, 1/2, P ).
a = b = c = 5/16.  */
multi_with_loss( E, With, P ) :-
	X pinl uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	X = a \\ With :: Y = x,
	X = b \\ With :: Y = x,
	P is p(X==E).

/* 
multi_rhs( a, 1/2, P ).
multi_rhs( b, 1/2, P ).
multi_rhs( c, 1/2, P ).
a = 31/96, b = 37/96, c = 7/24 */
multi_rhs( E, With, P ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([w,x,y,z]),
	Z pin uniform([s,t]),
	X = a \\ With :: Y = x,
	X = b \\ With :: Z = t,
	P is p(X==E).

/*
Currently the meta-interpreter does not check sum p_i =< 1.
pi_overdose( a, P ).
pi_overdose( b, P ).
pi_overdose( c, P ).
a = 7/18, b = 17/36, c = 2/9.  */
pi_overdose( Which, P ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([x,y,z]),
	X = a :: 1/2 ++ b = X :: 3/4 \\  Y = x,
	P is p(X==Which).

/*
pi_constricted( a, P ).
pi_constricted( b, P ).
pi_constricted( c, P ).
a = 7/18, b = 1/3, c = 2/9.  */
pi_constricted( Which, P ) :-
	X pin uniform([a,b,c]),
	Y pin uniform([x,y,z]),
	X = a :: 1/2 ++ b = X :: 1/3 \\  Y = x,
	P is p(X==Which).

%%% Demonstrate some available methods.

/* 
hpr_test( a, P ).
hpr_test( b, P ).
hpr_test( c, P ).
hpr_test( d, P ).
a = d = 7/62, b = c = 12/31. (sum = 1). */
hpr_test( X, P ) :-
	Y pin hypergeometric( [a,b,c,d], [16,16] ),
	P is p(X=Y).

%%% Some predicates which i havent checked yet
just_some_test( E, P ) :-
	X pin uniform([t,s,w]),
	Y pin uniform([w,x,y,z]),
	X \= W :: 1/1 \\ Y = W,
	X = t :: 1/3 ++ X = s :: 2/3 \\ Y = z,
	P is p(X==E).

:- demo.
